;****************
;*              *
;* SYSTEM CALLS *
;*              *
;****************

;	PROSIT_LOCK_HEADER	1

;---------------------------------------------------------------------
; Process_New(WORD filehandle)
;
; Function: Create a new process from the file with a specified handle
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Process_New:
prosit@0001:
	movem.l		a0-a2/a5-a6/d1-d7,-(sp)

	move.w		4+48(sp),d5		;Get handle of file off the USP

	TOGGLE_PROSIT

	bsr		RelocateFile		;Relocate program. a6.l = *file, a5.l = *_main
	tst.w		d0
	bne		Process_New_Fail
	addq.l		#6,a5			;Skip the DRD section

	move.w		$16(a6),d0		;Offset to export table
	move.w		2(a6,d0.w),d0		;Offset to exported data
	lea		0(a6,d0.w),a6		;a6.l = *Prosit data from file (skipping first word)

Process_New_From_Process_Fork:
	move.l		(a6)+,d7		;Process ID
	moveq		#0,d4
	move.w		(a6)+,d4		;Stack size
	moveq		#0,d3
	move.w		(a6)+,d3		;Message Buffer size

	bsr		CreateContextBlock	;d6.w = Context block handle

	move.w		d6,d0
	doorsos::DEREF	d6,a6
	move.w		d0,C_ID(a6)

	move.b		#0,-(sp)		;List item priority
	move.l		a6,-(sp)		;*List item
	move.l		#ContextList,-(sp)	;*List
	bsr		List_Insert
	add.w		#10,sp

Process_New_Fail:
	TOGGLE_PROSIT

	movem.l		(sp)+,a0-a2/a5-a6/d1-d7
	rts

;---------------------------------------------------------------------
; Process_Fork(BYTE *EntryPoint, BYTE *PrositData)
;
; Function: Fork an existing process
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Process_Fork:
prosit@0002:
	movem.l		a0-a2/a5-a6/d1-d7,-(sp)
	moveq		#0,d5
	move.l		4+48(sp),a5
	move.l		4+52(sp),a6

	TOGGLE_PROSIT
	bra		Process_New_From_Process_Fork


;---------------------------------------------------------------------
; Process_Terminate(WORD ID)
;
; Function: Terminate process
;
; Output: d0 = 0 if success, -1 if failure
;---------------------------------------------------------------------
Process_Terminate:
prosit@0003:
	movem.l		a0-a4/d1-d2/d6,-(sp)

	move.w		4+8*4(sp),d6	;ID
	TOGGLE_PROSIT

	move.w		d6,-(sp)
	move.l		#ContextList,-(sp)
	bsr		List_Find
	addq.l		#6,sp
	cmp.l		#0,a0
	beq		\done		;Process not found in list

;	PROSIT_LOCK	1

	move.l		LIST_NEXT(a0),a4	;This pointer will be removed in List_Remove!

	move.l		a0,-(sp)
	move.l		#ContextList,-(sp)
	bsr		List_Remove
	addq.l		#8,sp

	move.l		a0,a3

	move.w		C_PROG_HANDLE(a3),d0
	beq.s		\isfork			;Process was a fork and shouldn't be unrelocated
	move.l		($40),a0
	jsr		(a0)			;Unrelocate program

\isfork	move.l		C_PROSIT_DATARETURN(a3),a0
	clr.w		C_STATUS_FLAGS(a3)	;Set process' status to NOT_RUNNING and some other stuff
;	clr.l		PD_RETURN_EVENT(a0)		;Clear event
	clr.w		PD_MESSAGES(a0)			;Clear no. of messages
;	move.b		#PS_NOT_RUNNING,PD_STATUS(a3)	;Set process' status to NOT_RUNNING

	move.w		C_STACK_HANDLE(a3),-(sp)
	beq.s		\nostk
	jsr		doorsos::HeapFree	;De-allocate stack memory block

\nostk	move.w		C_MESSAGEBUFFER_HANDLE(a3),(sp)
	beq.s		\nombuf
	jsr		doorsos::HeapFree	;De-allocate stack memory block

\nombuf	move.w		d6,(sp)			;Not -(sp) because we've already used the stack!
	jsr		doorsos::HeapFree	;De-allocate context memory block
	addq.l		#2,sp

;	PROSIT_UNLOCK	1

	move.l		a4,a0
	cmp.w		CurrentProcessID,d6	;Self-terminated process?
	beq		SelectedNextProcess	;...yes, move on to the next one

	moveq		#0,d0			;Success

\done	TOGGLE_PROSIT

	movem.l		(sp)+,a0-a4/d1-d2/d6
	rts

;---------------------------------------------------------------------
; Process_Send_Message(WORD destination, WORD size, ...message...)
;
; Function: Send the <size> bytes from the stack to <destination>:s
;  message queue.
;
; Output: d0 =  0: success
;              -1: destination buffer full
;              -2: destination process isn't running
;---------------------------------------------------------------------
Process_Send_Message:
prosit@0004:
	movem.l		a0-a1/a6/d1-d3/d6-d7,-(sp)

	lea		4+32(sp),a6
	TOGGLE_PROSIT

	moveq		#-2,d0		;Assume "doesn't exist"

	move.w		(a6)+,d6	;Destination
;	bsr		List_FindP
	tst.w		d5
	blt.s		\done		:Process not found in list

	doorsos::DEREF	d6,a0

	move.w		C_MESSAGEBUFFER_HANDLE(a0),d1
	doorsos::DEREF	d1,a1		;*Message buffer
	move.w		(a1)+,d2	;Size of buffer
	move.w		(a1)+,d3	;Current buffer index
	sub.w		d3,d2		;Free space in buffer
	subq.w		#2,d2		;We also store the size of the message

	move.w		(a6)+,d1	;Size of message
	moveq		#-1,d0		;Assume "no space"
	cmp.w		d1,d2
	blt.s		\done		;Not enough free space in buffer

	add.w		#2,-2(a1)
	add.w		d1,-2(a1)
	lea		0(a1,d3.w),a1	;This is where we'll start writing our message
	subq.w		#1,d1
	move.w		d1,(a1)+	;Store size of message - 1
\sloop	move.b		(a6)+,(a1)+	;Copy message, byte-by-byte
	dbra.s		d1,\sloop

	move.l		C_PROSIT_DATARETURN(a0),a1
	add.b		#1,(a1)		;Increase program's "no. of messages" value
	moveq		#0,d0

\done	TOGGLE_PROSIT
	movem.l		(sp)+,a0-a1/a6/d1-d3/d6-d7
	rts

;---------------------------------------------------------------------
; Process_Read_Message(BYTE *buffer)
;
; Function: Copy the next message in the message queue into *buffer
;  and then remove it from the queue. NOTE: If a2.l = 0 the message
;  will not be copied but it will be removed.
;---------------------------------------------------------------------
Process_Read_Message:
prosit@0005:
	movem.l		a0-a2/d1-d3/d6-d7,-(sp)

	move.w		CurrentProcessID,d6
	TOGGLE_PROSIT
;	bsr		List_FindP
	TOGGLE_PROSIT

	tst.w		d5
	blt.s		\done		:Process not found in list

	doorsos::DEREF	d6,a0

	move.l		C_PROSIT_DATARETURN(a0),a1
	tst.w		(a1)
	beq.s		\done
	sub.b		#1,(a1)		;Decrease program's "no. of messages" value

	move.w		C_MESSAGEBUFFER_HANDLE(a0),d1
	doorsos::DEREF	d1,a1		;*Message buffer
	move.w		(a1)+,d2	;Size of buffer
	move.w		(a1)+,d3	;Current buffer index

	move.l		a1,a0		;*First message
	move.w		(a0)+,d1	;Size of message - 1
	sub.w		d1,d3
	move.l		4+32(sp),a2	;*buffer
	cmp.l		#0,a2
	beq.s		\skip

\sloop	move.b		(a0)+,(a2)+	;Copy message, byte-by-byte
	dbra.s		d1,\sloop

\skip	subq.w		#3,d3		;d3 = d3 - (size of message + 2)
	move.w		d3,-2(a1)
	subq.w		#1,d3
	blt.s		\done		;No remaining messages, we're done
\mloop	move.b		(a0)+,(a2)+	;Move remaining message buffer, byte-by-byte
	dbra.s		d3,\mloop

\done	movem.l		(sp)+,a0-a2/d1-d3/d6-d7
	rts

;---------------------------------------------------------------------
; Sleep(void)
;
; Function: Put process to sleep, i.e. it's moved from the list of
;  executed processes to the list of sleeping processes. It will
;  automatically wake up when a relevant event is triggered.
;---------------------------------------------------------------------
Sleep:
prosit@0006:
	TOGGLE_PROSIT

	move.l		a0,-(sp)
	move.l		SMCurrentProcess+4,a0
	movem.l		d0-d7/a0-a6,C_D0-C_D0(a0)	; Save D0-D7 and A0-A6
	move.l		(sp)+,C_A0-C_D0(a0)		;Restore process' a0

	move.l		usp,a1
	move.l		(a1),C_PC-C_D0(a0)	;PC = RTS adress
	addq.l		#4,a1
	move.l		a1,C_USP-C_D0(a0)	;Remove RTS adress from stack
	bset.b		#CS_SLEEPING,C_STATUS_FLAGS-C_D0(a0)

	move.b		#0,-(sp)
;	move.w		CurrentProcessID,-(sp)
	pea		-C_D0(a0)
	move.l		C_NEXT-C_D0(a0),a0	;Where the scheduler will pick things up
	move.l		#ContextList,-(sp)
	bsr		List_Remove		;Remove process from "running" list...
	move.l		#SleepingContextList,(sp)
	bsr		List_Insert		;...and put it in the "sleeping" list.
	add.w		#10,sp

	bra		SelectedNextProcess

;\back	movem.l		(sp)+,a0-a1
;	rts

;---------------------------------------------------------------------
; Wake_Up(WORD ID)
;
; Function: Wake up process <ID>, i.e. move it from the list of
;  sleeping processes to the list of executed processes. If the
;  process isn't sleeping nothing happens.
;---------------------------------------------------------------------
Wake_Up:
prosit@0007:
	movem.l		a0/d0,-(sp)
	move.w		4+2*4(sp),d0		;ID

	TOGGLE_PROSIT

	move.w		d0,-(sp)
	move.l		#SleepingContextList,-(sp)
	bsr		List_Find
	addq.l		#6,sp
	cmp.l		#0,a0
	beq		\done

	move.b		#0,-(sp)
;	move.w		LIST_ID(a0),-(sp)
	move.l		a0,-(sp)
	move.l		#SleepingContextList,-(sp)
	bsr		List_Remove
	move.l		#ContextList,(sp)
	bsr		List_Insert
	add.w		#10,sp
	bclr.b		#CS_SLEEPING,C_STATUS_FLAGS(a0)

\done	TOGGLE_PROSIT

	movem.l		(sp)+,a0/d0
	rts

;---------------------------------------------------------------------
; Trigger_Event(WORD processID, LONG event)
;
; Function: Execute the <event> subroutine inside <processID>.
;
; Output: d0.w >= 0 if success, -1 if failed
;---------------------------------------------------------------------
Trigger_Event:
prosit@0008:
	movem.l		a0-a3/d5-d7,-(sp)

	move.w		4+7*4(sp),d6		;processID

	move.w		d6,d5
	doorsos::DEREF	d5,a0
	move.l		4+7*4+2(sp),d5		;event

	move.l		C_PROSIT_DATARETURN(a0),a1
	lea		PD_NO_OF_EVENTS(a1),a2		;*Event table

	move.w		(a2)+,d7
	beq		\done			;Process has no events
	subq.w		#1,d7
\loop	cmp.l		(a2)+,d5
	beq.s		\found
	addq.l		#4,a2
	dbra.s		d7,\loop
	bra		\done			;Event not found

\found	btst.b		#CS_IN_INTERRUPT,C_STATUS_FLAGS(a0)
	beq.s		\ok			;No event is currently running, go on
	BREAK_AND_COOPERATE			;Event is running, wait for it to finish
	bra.s		\found

;Wait for process to return from an eventual library call
\ok
	move.w		C_PROG_HANDLE(a0),d7
	doorsos::DEREF	d7,a1
	move.l		a1,a3
	add.w		(a1),a3

\stick	cmp.l		C_PC(a0),a1
	bgt.s		\wait
	cmp.l		C_PC(a0),a3
	bgt.s		\trig
\wait	BREAK_AND_COOPERATE
	bra.s		\stick

\trig
	TOGGLE_PROSIT
	bset.b		#CS_IN_INTERRUPT,C_STATUS_FLAGS(a0)
	move.l		C_USP(a0),a1			;Get process' current SP
	move.l		C_PC(a0),-(a1)			;Push process' PC onto its stack
	move.w		C_SR(a0),-(a1)			;Push process' SR onto its stack
	move.l		#Return_From_Event,-(a1)	;Where we will return after an event
	move.l		a1,C_USP(a0)			;Store process' new SP
	move.l		(a2),C_PC(a0)			;Point process' PC to the event handler
	TOGGLE_PROSIT
	moveq		#0,d0
	btst.b		#CS_SLEEPING,C_STATUS_FLAGS(a0)
	beq.s		\done				;Process isn't sleeping
;Process is sleeping, wake it up
	move.w		d6,-(sp)
	bsr		Wake_Up
	addq.l		#2,sp
\done
	movem.l		(sp)+,a0-a3/d5-d7
	rts

Return_From_Event:
	TOGGLE_PROSIT
	subq.l		#6,sp			;Make room on the stack
	movem.l		a1/d0,-(sp)		;Store registers that are used here

	move.w		CurrentProcessID,d0
	doorsos::DEREF	d0,a1
	bclr.b		#CS_IN_INTERRUPT,C_STATUS_FLAGS(a1)

	move.l		usp,a1			;Process' SP
	move.w		(a1)+,8(sp)		;Process' SR
	move.l		(a1)+,8+2(sp)		;Process' PC
	move.l		a1,usp			;Retsore process' USP

	movem.l		(sp)+,a1/d0		;Restore used registers

	rte				;Return to process

;---------------------------------------------------------------------
; Broadcast_Event(LONG event)
;
; Function: Execute the <event> subroutine inside all processes
;  except the caller.
;---------------------------------------------------------------------
Broadcast_Event:
prosit@0009:
	movem.l		a0-a2/d0-d2,-(sp)

	move.l		4+6*4(sp),d2		;event

;Running processes
	move.l		ContextList,a0		;*First
	cmp.l		#0,a0
	beq.s		\sleep

\rloop	move.w		C_ID(a0),d0		;ProcessID
	move.l		C_NEXT(a0),a0
	cmp.w		CurrentProcessID,d0
	beq		\rnext			;The caller must not get the event!
	move.l		d2,-(sp)
	move.w		d0,-(sp)
	bsr		Trigger_Event
	addq.l		#6,sp
\rnext	cmp.l		ContextList,a0
	bne.s		\rloop			;Next process

;Sleeping processes
\sleep	move.l		#SleepingContextList,-(sp)
	bsr		List_Size
	addq.l		#4,sp
	subq.w		#1,d0
	blt.s		\done			;List is empty

	move.l		SleepingContextList,a0		;*First
	move.w		d0,d1
\sloop	move.w		C_ID(a0),d0		;ProcessID
	move.l		C_NEXT(a0),a0		;Get thsi first in case the proecss is waken up
	cmp.w		CurrentProcessID,d0
	beq		\snext			;The caller must not get the event!
	move.l		d2,-(sp)
	move.w		d0,-(sp)
	bsr		Trigger_Event
	addq.l		#6,sp
\snext	dbra.s		d1,\sloop		;Next process
;	cmp.l		SleepingContextList,a0
;	bne.s		\sloop			;Next process

\done	movem.l		(sp)+,a0-a2/d0-d2
	rts

;---------------------------------------------------------------------
; Get_File_Handle(BYTE *foldername, BYTE *filename)
;
; Function: Get the handle of a specific file
;
; Output: d5 = handle, 0 if file doesn't exist
;---------------------------------------------------------------------
Get_File_Handle:
prosit@000A:
	movem.l		a0-a4/d0-d1,-(sp)

	move.l		4+32(sp),a0
	move.l		4+28(sp),a1
	moveq		#0,d5		;In case file doesn't exist

;	TOGGLE_PROSIT

;Find folder
	move.l		doorsos::Heap,a2
	move.w		doorsos::FolderListHandle,d0
	lsl.w		#2,d0
	move.l		0(a2,d0.w),a2	;*Root
	addq.l		#2,a2
	move.w		(a2)+,d0	;Number of folders
	subq.w		#1,d0		;...-1
\loop1	move.l		a0,a3
	move.l		a2,a4
\cloop1	cmp.b		(a3)+,(a4)+
	bne.s		\next1
	tst.b		-1(a4)
	beq.s		\found1
	bra.s		\cloop1
\next1	lea		14(a2),a2	;Next entry
	dbra.s		d0,\loop1
	bra.s		\fail		;Folder not found

\found1	move.w		12(a2),d0
	doorsos::DEREF	d0,a2
	addq.l		#2,a2
	move.w		(a2)+,d0	;Number of files in folder
	subq.w		#1,d0		;...-1
\loop2	move.l		a1,a3
	move.l		a2,a4
\cloop2	cmp.b		(a3)+,(a4)+
	bne.s		\next2
	tst.b		-1(a4)
	beq.s		\found2
	bra.s		\cloop2
\next2	lea		14(a2),a2	;Next entry
	dbra.s		d0,\loop2
	bra.s		\fail		;File not found

\found2	move.w		12(a2),d5	;Found file, get handle

\fail	;TOGGLE_PROSIT

	movem.l		(sp)+,a0-a4/d0-d1
	rts

;---------------------------------------------------------------------
; Execute_File(WORD handle)
;
; Function: Execute a file.
; Output: d0.w = 0 if success, -1 if failed
;---------------------------------------------------------------------
Execute_File:
prosit@000B:
	movem.l		a0/a5/d5,-(sp)

	move.w		4+12(sp),d5

	bsr		RelocateFile
	tst.w		d0
	bne.s		\fail

	TOGGLE_PROSIT
	jsr		6(a5)		;Run file
	exg.w		d0,d5
	move.l		($40),a0
	jsr		(a0)		;Unrelocate file
	TOGGLE_PROSIT
	exg.w		d0,d5

\fail	movem.l		(sp)+,a0/a5/d5
	rts



















;PlotList:
;	movem.l		a0-a2/d7,-(sp)
;	move.l		4+4*4(sp),a0
;	move.w		4+4*4+4(sp),d7
;	HexShow		d7,#0,#0,#3
;	move.l		a0,a1
;\loop	HexShow		2(a0),d7,#1,#3
;	HexShow		C_PC(a0),d7,#6,#7
;	move.l		C_USP(a0),a2
;	HexShow		(a2),d7,#15,#7
;	addq.w		#1,d7
;	move.l		(a0),a0
;	cmp.l		a0,a1
;	bne.s		\loop
;	HexShow		#0,d7,#0,#7
;	HexShow		#0,d7,#8,#7
;\l	bra.s	\l
;	movem.l		(sp)+,a0-a2/d7
;	rts
