TOGGLE_TRAP		EQU	13
BREAK_COOPERATE_TRAP	EQU	14

PROSIT_HEADER	MACRO
		jmp		prosit@0000
		ENDM

BREAK_AND_COOPERATE	MACRO
		trap		#BREAK_COOPERATE_TRAP
		ENDM

PROSIT_LOCK_HEADER	MACRO	; Lockvar #
	P_Lock_\1:
		add.b		#1,p_lock_var_\1
		beq.s		P_Lock_Run_\1
		sub.b		#1,p_lock_var_\1
		BREAK_AND_COOPERATE
		bra.s		P_Lock_\1
	P_Lock_Run_\1:
		rts
	P_Unlock_\1:
		sub.b		#1,p_lock_var_\1
		rts
p_lock_var_\1	dc.w	-1
		ENDM

PROSIT_LOCK	MACRO	; Lockvar #
		bsr		P_Lock_\1
		ENDM

PROSIT_UNLOCK	MACRO	; Lockvar #
		bsr		P_Unlock_\1
		ENDM

TOGGLE_PROSIT	MACRO
		trap		#TOGGLE_TRAP
		ENDM

TRIGGER_EVENT	MACRO	; processID, event
		move.l		\2,-(sp)
		move.w		\1,-(sp)
		jsr		Prosit::Trigger_Event
		addq.l		#6,sp
		ENDM

BROADCAST_EVENT	MACRO	; event
		move.l		\1,-(sp)
		jsr		Prosit::Broadcast_Event
		addq.l		#4,sp
		ENDM

SEND_MESSAGE	MACRO	; DESTINATION ID, SIZE
		move.w		\2,-(sp)
		move.w		\1,-(sp)
		jsr		Prosit::Process_Send_Message
		add.l		\2+4,sp
		ENDM

;*********************************************************************
;*                       PROCESS STATUS MODES                        *
;*********************************************************************
PS_NOT_RUNNING		EQU	0
PS_RUNNING		EQU	1
PS_SLEEPING		EQU	2


;*********************************************************************
;*                         SYSTEM FUNCTIONS                          *
;*********************************************************************

;---------------------------------------------------------------------
; Process_New(WORD filehandle)
;
; Function: Create a new process from the file with a specified handle
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_New		equ	prosit@0001

;---------------------------------------------------------------------
; Process_Fork(BYTE *EntryPoint, BYTE *PrositData)
;
; Function: Fork an existing process
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_Fork		equ	prosit@0002

;---------------------------------------------------------------------
; Process_Terminate(WORD ID)
;
; Function: Terminate process
;
; Output: d0 = 0 if success, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_Terminate	equ	prosit@0003

;---------------------------------------------------------------------
; Process_Send_Message(WORD destination, WORD size, ...message...)
;
; Function: Send the <size> bytes from the stack to <destination>:s
;  message queue.
;
; Output: d0 =  0: success
;              -1: destination buffer full
;              -2: destination process isn't running
;---------------------------------------------------------------------
Prosit::Process_Send_Message	equ	prosit@0004

;---------------------------------------------------------------------
; Process_Read_Message(BYTE *buffer)
;
; Function: Copy the next message in the message queue into *buffer
;  and then remove it from the queue. NOTE: If a2.l = 0 the message
;  will not be copied but it will be removed.
;---------------------------------------------------------------------
Prosit::Process_Read_Message	equ	prosit@0005

;---------------------------------------------------------------------
; Sleep(void)
;
; Function: Put process to sleep, i.e. it's moved from the list of
;  executed processes to the list of sleeping processes. It will
;  automatically wake up when a relevant event is triggered.
;---------------------------------------------------------------------
Prosit::Sleep			equ	prosit@0006

;---------------------------------------------------------------------
; Wake_Up(WORD ID)
;
; Function: Wake up process <ID>, i.e. move it from the list of
;  sleeping processes to the list of executed processes. If the
;  process isn't sleeping nothing happens.
;---------------------------------------------------------------------
Prosit::Wake_Up			equ	prosit@0007

;---------------------------------------------------------------------
; Trigger_Event(WORD processID, LONG event)
;
; Function: Execute the <event> subroutine inside <processID>.
;
; Output: d0.w = 0 if success, -1 if failed
;---------------------------------------------------------------------
Prosit::Trigger_Event		equ	prosit@0008

;---------------------------------------------------------------------
; Broadcast_Event(LONG event)
;
; Function: Execute the <event> subroutine inside all processes
;  except the caller.
;---------------------------------------------------------------------
Prosit::Broadcast_Event		equ	prosit@0009

;---------------------------------------------------------------------
; Get_File_Handle(BYTE *foldername, BYTE *filename)
;
; Function: Get the handle of a specific file
;
; Output: d5 = handle, 0 if file doesn't exist
;---------------------------------------------------------------------
Prosit::Get_File_Handle		equ	prosit@000A

;---------------------------------------------------------------------
; Execute_File(WORD handle)
;
; Function: Execute a file.
;
; Output: d0.w = 0 if success, -1 if failed
;---------------------------------------------------------------------
Prosit::Execute_File		equ	prosit@000B

;*********************************************************************
;*                         SYSTEM VARIABLES                          *
;*********************************************************************

;---------------------------------------------------------------------
;WORD: CurrentProcessID
;Current process' ID (for subroutines and others)
;---------------------------------------------------------------------
Prosit::CurrentProcessID	equ	prosit@000C

;---------------------------------------------------------------------
;LONG: Ticks
;Number of ticks since Prosit was started. Is contained in a longword
;so it restarts after a while, but by then the batteries will probably
;be dead :)
;---------------------------------------------------------------------
Prosit::Ticks			equ	prosit@000D

;---------------------------------------------------------------------
;BYTE: ControlKeys
;The mask of the row in the keyboard matrix that contains the arrow-
;and shiftkeys:
; Bit 0: 2nd
; Bit 1: Diamond
; Bit 2: Shift
; Bit 3: Hand
; Bit 4: Left
; Bit 5: Up
; Bit 6: Right
; Bit 7: Down
;---------------------------------------------------------------------
Prosit::ControlKeys		equ	prosit@000E

;---------------------------------------------------------------------
;WORD: CurrentKey
;Holds the keycode of the current key being held down, excluding the
;arrow keys.
;---------------------------------------------------------------------
Prosit::CurrentKey		equ	prosit@000F

;---------------------------------------------------------------------
;WORD: MouseX
;Current mouse X coordinate
;---------------------------------------------------------------------
Prosit::MouseX			equ	prosit@0010

;---------------------------------------------------------------------
;WORD: MouseY
;Current mouse Y coordinate
;---------------------------------------------------------------------
Prosit::MouseY			equ	prosit@0011



;*********************************************************************
;*                     LIST HEADERS AND FUNCTIONS                    *
;*********************************************************************
LIST_NEXT		EQU	0
LIST_ID			EQU	LIST_NEXT+4
LIST_PRIORITY		EQU	LIST_ID+2
LIST_BLANK		EQU	LIST_PRIORITY+1
LIST_HEADER_SIZE	EQU	LIST_BLANK+1

;---------------------------------------------------------------------
; List_Insert(BYTE *list, WORD ID, BYTE *structure)
;
; Function: Insert <structure> first in <list>.
;---------------------------------------------------------------------
Prosit::List_Insert		equ	prosit@0012

;---------------------------------------------------------------------
; List_Remove(BYTE *list, BYTE *structure)
;
; Function: Remove <structure> from <list>.
;---------------------------------------------------------------------
Prosit::List_Remove		equ	prosit@0013

;---------------------------------------------------------------------
; List_Previous(BYTE *structure)
;
; Function: Get a pointer to the structure in <list> that points to
; <structure>
;
; Output: a0.l = *previous structure
;---------------------------------------------------------------------
Prosit::List_Previous		equ	prosit@0014

;---------------------------------------------------------------------
; List_Find(BYTE *list, WORD ID)
;
; Function: Get a pointer to the structure in <list> with <ID>.
;
; Output: a0.l = *structure, 0 if not found
;---------------------------------------------------------------------
Prosit::List_Find		equ	prosit@0015

;---------------------------------------------------------------------
; List_Size(BYTE *list)
;
; Function: Get number of elements in list.
;
; Output: d0.w = No. of elements in list.
;---------------------------------------------------------------------
Prosit::List_Size		equ	prosit@0016
