;---------------------------------------------------------------------
; Clear(WORD ID)
;
; Function: Clear the window contents block.
;---------------------------------------------------------------------
Clear:
winlib@000A:
	movem.l		a0/d0-d1,-(sp)

	move.w		12+4(sp),d0			;Get ID (handle) of window
	doorsos::DEREF	d0,a0
	move.w		W_WIDTH_IN_WORDS(a0),d0		;Width in words
	mulu.w		W_HEIGHT(a0),d0			;Width*Height in words
	subq.w		#1,d0
	move.w		W_CONTENTS_HANDLE(a0),d1	;Contents handle
	doorsos::DEREF	d1,a0				;*Contents

\cloop	clr.w		(a0)+				;Clear contents memory block
	dbra.s		d0,\cloop

	movem.l		(sp)+,a0/d0-d1
	rts

;---------------------------------------------------------------------
; SetClip(WORD ID, BYTE *clip)
;
; Function: Set window <ID> to use the clip area <clip>.
;---------------------------------------------------------------------
SetClip:
winlib@000B:
	movem.l		a0-a1/d0-d1,-(sp)

	move.w		4+4*4+0(sp),d0
	doorsos::DEREF	d0,a0		;*window
	move.l		4+4*4+2(sp),a1	;*clip

	move.w		(a1),d0		;x1
	move.w		4(a1),d1	;x2
	cmp.w		d1,d0
	ble.s		\xok
	exg.w		d1,d0		;x1<x2
\xok	tst.w		d0
	bge.s		\x1pos
	moveq		#0,d0		;x1 must be positive
\x1pos	cmp.w		W_WIDTH(a0),d1
	blt.s		\x2ok
	move.w		W_WIDTH(a0),d1
	subq.w		#1,d1		;x2 must be within the window!
\x2ok	move.w		d0,W_LEFT_CLIP(a0)
	move.w		d1,W_RIGHT_CLIP(a0)

	move.w		2(a1),d0	;y1
	move.w		6(a1),d1	;y2
	cmp.w		d1,d0
	ble.s		\yok
	exg.w		d1,d0		;y1<y2
\yok	tst.w		d0
	bge.s		\y1pos
	moveq		#0,d0		;y1 must be positive
\y1pos	cmp.w		W_HEIGHT(a0),d1
	blt.s		\y2ok
	move.w		W_HEIGHT(a0),d1
	subq.w		#1,d1		;y2 must be within the window!
\y2ok	move.w		d0,W_TOP_CLIP(a0)
	move.w		d1,W_BOTTOM_CLIP(a0)

\done	movem.l		(sp)+,a0-a1/d0-d1
	rts


;---------------------------------------------------------------------
; VertLine(WORD ID, WORD x, WORD y1, WORD y2, WORD flags,
;  [WORD pattern])
;
; Function: Draw a line from (x,y1) to (x,y2).
;
; Flags: Bit 0 = Pattern/no pattern
;        Bit 1 = Color (if no pattern)
;---------------------------------------------------------------------
VertLine:
winlib@000C:
	movem.l		a0/d0-d6,-(sp)

	move.w		4+8*4(sp),d0	;Window handle
	doorsos::DEREF	d0,a0		;*Window

	move.w		4+8*4+2(sp),d0	;x
	cmp.w		W_LEFT_CLIP(a0),d0
	blt		\done		;outside left clip
	cmp.w		W_RIGHT_CLIP(a0),d0
	bgt		\done		;outside right clip

	move.w		4+8*4+4(sp),d1	;y1
	move.w		4+8*4+6(sp),d2	;y2
	move.w		4+8*4+8(sp),d4	;Flags and pattern

	cmp.w		d1,d2
	bge.s		\ok
	exg.w		d1,d2		;Make sure y1<=y2

\ok	cmp.w		W_BOTTOM_CLIP(a0),d1
	bgt		\done		;Entire line below clip
	cmp.w		W_TOP_CLIP(a0),d1
	bge.s		\y1ok
	move.w		W_TOP_CLIP(a0),d1
\y1ok	cmp.w		W_TOP_CLIP(a0),d2
	blt		\done		;Entire line above clip
	cmp.w		W_BOTTOM_CLIP(a0),d2
	ble.s		\y2ok
	move.w		W_BOTTOM_CLIP(a0),d2
\y2ok	sub.w		d1,d2		;Height-1
	move.w		W_WIDTH_IN_WORDS(a0),d6
	lsl.w		#1,d6		;Bytes per row

	move.w		W_CONTENTS_HANDLE(a0),d3
	doorsos::DEREF	d3,a0
	move.w		d0,d3
	lsr.w		#3,d3
	mulu.w		d6,d1
	add.w		d1,d3
	lea		0(a0,d3.w),a0	;*start plot

	move.w		d0,d3
	and.w		#%0000000000000111,d3
	neg.w		d3
	add.w		#7,d3		;Bit to affect in row

	btst.b		#LB_USES_PATTERN,d4
	beq.s		\normal

;* LINE WITH PATTERN *
	moveq		#-1,d4
	moveq		#0,d5
	move.w		4+8*4+10(sp),d5	;Pattern
\flen	addq.w		#1,d4
	btst.b		d4,d5
	beq.s		\flen
	neg.w		d4
	add.w		#14,d4		;Length of pattern - 1
	move.w		d4,d1

\loop	bclr.b		d3,(a0)
	btst.w		#15,d5
	beq.s		\clr
	bset.b		d3,(a0)
\clr	rol.l		#1,d5
	dbra.s		d1,\nxtbit
	ror.l		d4,d5
	ror.l		#1,d5
	move.w		d4,d1
\nxtbit	add.w		d6,a0
	dbra.s		d2,\loop
	bra.s		\done

;* NORMAL LINE *
\normal	btst.b		#LB_COLOR,d4
	beq.s		\white
\black	bset.b		d3,(a0)		;Color = Black
	add.w		d6,a0
	dbra.s		d2,\black
	bra.s		\done

\white	bclr.b		d3,(a0)		;Color = White
	add.w		d6,a0
	dbra.s		d2,\white

\done	movem.l		(sp)+,a0/d0-d6
	rts


;---------------------------------------------------------------------
; HorizLine(WORD ID, WORD y, WORD x1, WORD x2, WORD flags,
;  [WORD pattern])
;
; Function: Draw a line from (x1,y) to (x2,y).
;
; Flags: Bit 0 = Pattern/no pattern
;        Bit 1 = Color (if no pattern)
;---------------------------------------------------------------------
HorizLine:
winlib@000D:
	movem.l		a0/a2/d0-d7,-(sp)

	move.w		4+10*4(sp),d0	;Window handle
	doorsos::DEREF	d0,a0		;*Window

	move.w		4+10*4+2(sp),d0	;y
	cmp.w		W_TOP_CLIP(a0),d0
	blt		\done		;outside upper clip
	cmp.w		W_BOTTOM_CLIP(a0),d0
	bgt		\done		;outside lower clip

	move.w		4+10*4+4(sp),d1	;x1
	move.w		4+10*4+6(sp),d2	;x2
	move.w		4+10*4+8(sp),d4	;Flags and pattern

	cmp.w		d1,d2
	bge.s		\ok
	exg.w		d1,d2		;Make sure x1<=x2

\ok	cmp.w		W_RIGHT_CLIP(a0),d1
	bgt		\done		;Entire line to the right of clip
	cmp.w		W_LEFT_CLIP(a0),d1
	bge.s		\x1ok
	move.w		W_LEFT_CLIP(a0),d1
\x1ok	cmp.w		W_LEFT_CLIP(a0),d2
	blt		\done		;Entire line to the left of clip
	cmp.w		W_RIGHT_CLIP(a0),d2
	ble.s		\x2ok
	move.w		W_RIGHT_CLIP(a0),d2
\x2ok	mulu.w		W_WIDTH_IN_WORDS(a0),d0
	lsl.w		#1,d0		;Bytes per row

	move.w		W_CONTENTS_HANDLE(a0),d3
	doorsos::DEREF	d3,a0		;*Contents
	move.w		d1,d3
	lsr.w		#3,d3
	bclr.b		#0,d3		;We're working with words
	add.w		d0,d3
	lea		0(a0,d3.w),a0	;*start plot

	lea		wmasks(pc),a2

	move.w		d2,d3
	and.w		#%0000000000001111,d3	;End bit
	lsl.w		#1,d3
	move.w		32(a2,d3.w),d7		;Right mask
	lsr.w		#4,d2			;Word of right edge

	move.w		d1,d3
	and.w		#%0000000000001111,d3	;Start bit
	lsl.w		#1,d3
	move.w		0(a2,d3.w),d6		;Left mask

	move.w		d1,d3
	lsr.w		#4,d3			;Word of left edge
	sub.w		d3,d2
	subq.w		#1,d2			;No. of words in-between - 1
	bge.s		\isok

	and.w		d7,d6			;Combine masks since only one word is showing
	moveq		#0,d7

\isok	btst.b		#LB_USES_PATTERN,d4
	beq		\normal

;* LINE WITH PATTERN *
	moveq		#-1,d4
	moveq		#0,d5
	move.w		4+10*4+10(sp),d5	;Pattern
\flen	addq.w		#1,d4
	btst.w		d4,d5
	beq.s		\flen
	bclr.w		d4,d5
	neg.w		d4
	add.w		#15,d4		;Length of pattern
	swap		d5		;Put pattern in upper word

;Create the pattern longword
	moveq		#0,d0
	move.w		#32,d3
\make	lsr.l		d4,d0
	or.l		d5,d0
	sub.w		d4,d3
	bgt.s		\make
	swap		d0

	and.w		#%0000000000001111,d1	;RS = x mod 16

	move.w		d0,d5
	lsr.w		d1,d5
	and.w		d6,d5
	not.w		d6
	move.w		(a0),d3
	and.w		d6,d3
	or.w		d5,d3
	move.w		d3,(a0)+		;First word

	neg.w		d1
	add.w		#16,d1			;First LS = 16-RS

	moveq		#16,d3
\make2	sub.w		d4,d3
	bge.s		\make2
	add.w		d4,d3			;16 mod len

	subq.w		#1,d2
	blt.s		\end

\loop	move.l		d0,d5
	rol.l		d1,d5
	move.w		d5,(a0)+
	add.w		d3,d1
	cmp.w		d4,d1
	blt.s		\nxt
	sub.w		d4,d1
\nxt	dbra.s		d2,\loop

\end	rol.l		d1,d0
	and.w		d7,d0
	not.w		d7
	move.w		(a0),d3
	and.w		d7,d3
	or.w		d0,d3
	move.w		d3,(a0)

	bra.s		\done

;* NORMAL LINE *
\normal	btst.b		#LB_COLOR,d4
	beq.s		\white

	or.w		d6,(a0)+		;Color = Black
	subq.w		#1,d2
	blt.s		\bnend
\bnloop	move.w		#$FFFF,(a0)+
	dbra.s		d2,\bnloop
\bnend	or.w		d7,(a0)
	bra.s		\done

\white	not.w		d6			;Color = White
	and.w		d6,(a0)+
	subq.w		#1,d2
	blt.s		\wnend
\wnloop	clr.w		(a0)+
	dbra.s		d2,\wnloop
\wnend	not.w		d7
	and.w		d7,(a0)

\done	movem.l		(sp)+,a0/a2/d0-d7
	rts

;---------------------------------------------------------------------
; Box(WORD ID, WORD x, WORD y, WORD width, WORD height, WORD flags,
;  [WORD pattern], [WORD pattern], [WORD pattern], [WORD pattern],
;  [BYTE *gfx, BYTE *mask]
;
; Function: Draw a <width,height>-sized box at (x, y).
; Flags:  Bit 0 = Top frame / no top frame
;         Bit 1 = Top pattern / no top pattern
;         Bit 2 = Top color (if no pattern)
;         Bit 3 = Bottom frame / no top frame
;         Bit 4 = Bottom pattern / no top pattern
;         Bit 5 = Bottom color (if no pattern)
;         Bit 6 = Left frame / no top frame
;         Bit 7 = Left pattern / no top pattern
;         Bit 8 = Left color (if no pattern)
;         Bit 9 = Right frame / no top frame
;        Bit 10 = Right pattern / no top pattern
;        Bit 11 = Right color (if no pattern)
;        Bit 12 = Filled / not filled
;        Bit 13 = Fill pattern / no fill pattern
;        Bit 14 = Fill color (if no pattern)
;                 Choices for stretch/center/tile?
;---------------------------------------------------------------------
Box:
winlib@000E:
	movem.l		a0-a6/d0-d7,-(sp)

	move.w		4+15*4+0(sp),d0		;Window Handle
	move.w		d0,d7			;Store it
	doorsos::DEREF	d0,a0			;*Window
	move.w		W_CONTENTS_HANDLE(a0),d0
	doorsos::DEREF	d0,a2			;*Contents

	move.w		4+15*4+2(sp),d0		;x1
	cmp.w		W_RIGHT_CLIP(a0),d0
	bgt		\done
	move.w		4+15*4+4(sp),d1		;y1
	cmp.w		W_BOTTOM_CLIP(a0),d1
	bgt		\done
	move.w		4+15*4+6(sp),d2		;width
	add.w		d0,d2
	subq.w		#1,d2			;x2
	cmp.w		W_LEFT_CLIP(a0),d2
	blt		\done
	move.w		4+15*4+8(sp),d3	;height
	add.w		d1,d3
	subq.w		#1,d3			;y2
	cmp.w		W_TOP_CLIP(a0),d3
	blt		\done

	move.w		4+15*4+10(sp),d4	;Flags

	ror.w		#1,d4

	lea		4+15*4+12(sp),a1

	sub.w		#12,sp
	move.w		d7,(sp)
;Top line
	move.w		d2,6(sp)	;x2
	move.w		d0,4(sp)	;x1
	btst.w		#15,d4		;Draw line?
	beq.s		\bot
	btst.w		#0,d4		;Use pattern?
	beq.s		\ntpat
	move.w		(a1)+,10(sp)	;Pattern
\ntpat	move.w		d4,8(sp)	;Flags
	move.w		d1,2(sp)	;y1
	bsr		HorizLine
	addq.w		#1,d1
;Bottom line
\bot	ror.w		#3,d4
	btst.w		#15,d4		;Draw line?
	beq.s		\left
	btst.w		#0,d4		;Use pattern?
	beq.s		\nbpat
	move.w		(a1)+,10(sp)	;Pattern
\nbpat	move.w		d4,8(sp)	;Flags
	move.w		d3,2(sp)	;y2
	bsr		HorizLine
	subq.w		#1,d3
;Left line
\left	move.w		d3,6(sp)	;y2
	move.w		d1,4(sp)	;y1
	ror.w		#3,d4
	btst.w		#15,d4		;Draw line?
	beq.s		\right
	btst.w		#0,d4		;Use pattern?
	beq.s		\nlpat
	move.w		(a1)+,10(sp)	;Pattern
\nlpat	move.w		d4,8(sp)	;Flags
	move.w		d0,2(sp)	;x1
	bsr		VertLine
	addq.w		#1,d0
;Right line
\right	ror.w		#3,d4
	btst.w		#15,d4		;Draw line?
	beq.s		\donlin
	btst.w		#0,d4		;Use pattern?
	beq.s		\nrpat
	move.w		(a1)+,10(sp)	;Pattern
\nrpat	move.w		d4,8(sp)	;Flags
	move.w		d2,2(sp)	;x2
	bsr		VertLine
	subq.w		#1,d2

\donlin	add.w		#12,sp

;*******************
;* INTERNAL OF BOX *
;*******************
	btst.w		#12-10,d4
	beq		\done		;No fill

;**************
;* RIGHT MASK *
;**************
\xdone	lea		wmasks(pc),a1
	cmp.w		W_RIGHT_CLIP(a0),d2
	ble.s		\xpos
;Right clipping
	move.w		W_RIGHT_CLIP(a0),d2
;No right clipping
\xpos	move.w		d2,d7
	and.w		#%0000000000001111,d7
	lsl.w		#1,d7
	move.w		32(a1,d7.w),d7		;Right edge mask
;*************
;* LEFT MASK *
;*************
	cmp.w		W_LEFT_CLIP(a0),d0
	bge.s		\x2pos
;Left clipping
	move.w		W_LEFT_CLIP(a0),d0
;No left clipping
\x2pos	move.w		d0,d6
	and.w		#%0000000000001111,d6
	lsl.w		#1,d6
	move.w		0(a1,d6.w),d6		;Left edge mask

;*******************
;* CALCULATE WIDTH *
;*******************
	lsr.w		#3,d0		;x1/8
	bclr.b		#0,d0		;We're plotting words
	add.w		d0,a2		;Move *plot, x-wise
	lsr.w		#1,d0
	lsr.w		#4,d2		;x2/16
	sub.w		d0,d2
	bgt.s		\wpos
;Left and right edge are in the same place...
	and.w		d7,d6
	moveq		#0,d7

\wpos	move.w		W_WIDTH_IN_WORDS(a0),a1
	add.w		a1,a1		;RowAdd for window contents
	subq.w		#1,d2		;Words to plot - 1

;*******************************************************
;* CALCULATE WHERE TO START PLOTTING AND HOW MANY ROWS *
;*******************************************************
	move.w		d3,d5
	sub.w		d1,d5			;y2-y1 = height-1

	move.w		W_TOP_CLIP(a0),d0
	sub.w		d1,d0			;y1clip-y1
	ble.s		\oktop
;Top is clipped, adjust parameters
	sub.w		d0,d5			;Subtract difference from height
	add.w		d0,d1			;Add difference to y1 => y1=y1clip
;Top isn't clipped
\oktop	move.w		d3,d0
	sub.w		W_BOTTOM_CLIP(a0),d0	;y2-y2clip
	ble.s		\okbot
;Bottom is clipped, adjust parameters
	sub.w		d0,d5			;Subtract difference from height
;Bottom isn't clipped
\okbot	move.w		a1,d0
	mulu.w		d1,d0
	add.w		d0,a2			;Move *plot, y-wise

;d2.w = No. of middle segments
;d4.w = Flags, right-rotated 10 bits
;d5.w = Height
;d6.w = Left mask
;d7.w = Right mask
;a0 = *window
;a1 = RowAdd for window contents
;a2 = *plot

	move.w		d6,d1
	move.w		d7,d3
	not.w		d6
	not.w		d7
	move.w		#$FFFF,\tploop+2
	btst.w		#14-10,d4		;Check color
	bne.s		\row			;Fill color = black
	moveq		#0,d1
	moveq		#0,d3
	move.w		#$0000,\tploop+2

;*********************
;* PLOT INTERNAL BOX *
;*********************
\row	move.l		a2,a3
	and.w		d6,(a3)
	or.w		d1,(a3)+	;Left edge
	move.w		d2,d0
	blt.s		\tpdone		;Left edge is combined, done with row
	beq.s		\tplast		;No middle segments, plot right edge
	subq.w		#1,d0
\tploop	move.w		#$FFFF,(a3)+	;Plot middle segments
	dbra.s		d0,\tploop
\tplast	and.w		d7,(a3)		;Right edge
	or.w		d3,(a3)
\tpdone	add.w		a1,a2
	dbra.s		d5,\row

\done	movem.l		(sp)+,a0-a6/d0-d7
	rts


;---------------------------------------------------------------------
; PlotGFX(WORD ID, WORD x, WORD y, BYTE *gfx, BYTE *mask)
;
; Function: Plot <gfx> with mask <mask> at (x,y) in window <ID>, using
;  the window's current clip area. The format for gfx is described in
;  the docs.
;---------------------------------------------------------------------
PlotGFX:
winlib@000F:
	movem.l		a0-a6/d0-d7,-(sp)

	move.w		4+15*4+0(sp),d0		;Window Handle
	doorsos::DEREF	d0,a3			;*Window
	move.w		4+15*4+2(sp),d0		;x
	move.w		4+15*4+4(sp),d1		;y
	move.l		4+15*4+6(sp),a0		;*gfx
	move.l		4+15*4+10(sp),a6	;*mask

	move.w		(a0)+,d2		;gfx_width (bytes)
	move.w		(a0)+,d3		;gfx_height (pixels)
	move.w		d2,a4			;Store gfx_width for later use

	cmp.w		W_RIGHT_CLIP(a3),d0
	bgt		\exit			;Plot is to the right of clip, exit
	move.w		d2,d6
	lsl.w		#3,d6
	add.w		d0,d6
	cmp.w		W_LEFT_CLIP(a3),d6
	ble		\exit			;Plot is to the left of clip, exit
	cmp.w		W_BOTTOM_CLIP(a3),d1
	bgt		\exit			;Plot is below clip, exit
	move.w		d1,d7
	add.w		d3,d7
	cmp.w		W_TOP_CLIP(a3),d7
	ble		\exit			;Plot is above clip, exit

	move.w		W_CONTENTS_HANDLE(a3),d6
	doorsos::DEREF	d6,a2
	move.w		W_TOP_CLIP(a3),d6
	mulu.w		W_WIDTH_IN_WORDS(a3),d6
	lsl.w		#1,d6			;Since window width is in words
	move.w		W_LEFT_CLIP(a3),d7
	lsr.w		#3,d7
	add.w		d7,d6
	add.w		d6,a2			;*plot

;****************
;* Y PARAMETERS *
;****************
;Lower Y edge
	move.w		d1,d6
	add.w		d3,d6			;y + gfx_height
	subq.w		#1,d6
	sub.w		W_BOTTOM_CLIP(a3),d6		;y + gfx_height - y2_clip
	blt.s		\cont1
;Lower edge of plot exceeds lower clip limit, decrease no. of rows to plot
	sub.w		d6,d3

;Upper Y edge
\cont1	move.w		d1,d6
	sub.w		W_TOP_CLIP(a3),d6	;y - y1_clip
	bge.s		\ypos
;GFX y starts before the clip, adjust GFX start parameters
	add.w		d6,d3			;Decrease no. of rows to plot
	muls.w		d2,d6			;(y - y1_clip)*gfx_width
	sub.w		d6,a0			;Move *gfx
	sub.w		d6,a6			;Move *mask
	bra.s		\cont2
;GFX y starts on or after the clip, adjust destination start parameters
\ypos	mulu.w		W_WIDTH_IN_WORDS(a3),d6	;(y - y1_clip)*target_width
	lsl.w		#1,d6			;Since window wisth is in words
	add.w		d6,a2			;Move *plot

;
; Now d3 = No. of rows to plot
;     a0 = Where to start getting data, y-wise
;     a2 = Where to start plotting, y-wise
;

\cont2

;************************************
;* CREATE LEFT- AND RIGHTCLIP MASKS *
;************************************
;Create left bitmask
	move.w		W_LEFT_CLIP(a3),d7
	cmp.w		d0,d7
	bge.s		\oklc
	move.w		d0,d7
\oklc	and.w		#%0000000000000111,d7
	lea		masks(pc),a5
	move.b		0(a5,d7.w),d6		;d6.uw = left mask
	swap		d6

;Create right bitmask
	move.w		d0,d7
	move.w		a4,d6
	lsl.w		#3,d6
	subq.w		#1,d7
	add.w		d6,d7			;x+width
	cmp.w		W_RIGHT_CLIP(a3),d7
	ble.s		\okrc
	sub.w		#8,a5
	move.w		W_RIGHT_CLIP(a3),d7
\okrc	and.w		#%0000000000000111,d7
	move.b		16(a5,d7.w),d6		;d6.lw = right mask

;****************
;* X PARAMETERS *
;****************
;Right X edge
	move.w		d0,d7
	asr.w		#3,d7			;x/8

	move.w		W_RIGHT_CLIP(a3),d4
	lsr.w		#3,d4			;x2_clip/8
	sub.w		d7,d4			;(x2_clip/8) - (x/8)
	beq.s		\sp1

	move.w		#$4E71,\sm2		;NOP
	move.w		#$4E71,\sm21		;NOP
	cmp.w		d4,d2
	blt.s		\cont3
	bne.s		\ne
	subq.w		#1,d4

;(x/8)+width is on or to the right of (x2_clip/8)
\ne	move.w		#$1210,\sm2		;move.b (a0),d1
	move.w		#$1A16,\sm21		;move.b (a6),d5
	move.w		d4,d2
	bra.s		\cont3

;(x/8) is on the right clip byte, alter the masks and set the width to handle it
\sp1	moveq		#1,d2		;Set width to one to "fool" plot routine
	and.b		d6,d5		;Combine left and right masks into the left mask...
	move.b		#0,d6		;...and clear the old right mask

;Left X edge
\cont3	move.w		W_LEFT_CLIP(a3),d4
	lsr.w		#3,d4			;x1_clip/8
	sub.w		d7,d4			;(x1_clip/8) - (x/8)
	ble.s		\xpos2
;(x/8) is to the left of (x1_clip/8), adjust GFX start parameters
	add.w		d4,a0			;Move *gfx
	add.w		d4,a6			;Move *mask
	sub.w		d4,d2			;Decrease no. of columns to plot
	bne.s		\nosp2

	moveq		#1,d2		;Set width to one to "fool" plot routine
	and.b		d5,d6		;Combine left and right masks into the right mask...
	move.b		#0,d5		;...and clear the old left mask
	sub.w		#1,a2		;Move *plot...
	subq.w		#1,a0		;...and *gfx. Necessary since this is a special case.
	subq.w		#1,a6		;...and *mask.

\nosp2	move.l		#$1228FFFF,\sm1		;Self-modifying code: use a PreByte
	move.l		#$1A2EFFFF,\sm11
	bra.s		\cont4
;(x/8) is on or to the right of (x1_clip/8), adjust destination start parameters
\xpos2	sub.w		d4,a2			;Move *plot
	move.l		#$123C0000,\sm1		;Self-modifying code: don't use a PreByte
	move.l		#$1A3C0000,\sm11

\cont4
	sub.w		d2,a4			;Source RowAdd
	move.w		W_WIDTH_IN_WORDS(a3),a5
	add.w		a5,a5			;Since window width is in words
	sub.w		d2,a5			;Target RowAdd

	subq.w		#1,d3			;We're using dbra.
	subq.w		#2,d2			;We're using dbra, plus one byte will be used outside the loop.

	and.w		#%0000000000000111,d0

;******************
;*                *
;* START PLOTTING *
;*                *
;******************
	move.w		d2,a1

\row	swap		d6			;Swap in left mask into d6.lw

\sm1	move.b		-1(a0),d1		;Self-modifying code: first byte to use before plotting
\sm11	move.b		-1(a6),d5		;Self-modifying code: first byte to use before plotting

	lsl.w		#8,d1		;Prepare for first gfx
	lsl.w		#8,d5		;Prepare for first mask
	move.b		(a0)+,d1	;First gfx on row
	move.b		(a6)+,d5	;First mask on row
	move.w		d1,d4		;Store gfx
	move.w		d5,d2		;Store mask
	lsr.w		d0,d4		;Shift gfx
	lsr.w		d0,d2		;Shift mask
	and.b		d6,d2		;Mask + left mask
	and.b		d2,d4		;Mask first gfx

	move.b		(a2),d7		;LCD
	not.b		d2		;Invert mask
	and.b		d2,d7		;Mask LCD
	or.b		d7,d4		;LCD + new gfx

	move.b		d4,(a2)+	;Plot

	move.w		a1,d7
	blt.s		\nxt
\col	lsl.w		#8,d1		;Prepare for next gfx
	lsl.w		#8,d5		;Prepare for next mask
	move.b		(a0)+,d1	;Next gfx
	move.b		(a6)+,d5	;Next mask

	move.w		d1,d4		;Store gfx
	move.w		d5,d2		;Store mask
	lsr.w		d0,d4		;Shift gfx
	lsr.w		d0,d2		;Shift mask
	and.b		d2,d4		;Mask gfx
	not.b		d2		;Invert mask
	and.b		(a2),d2		;Masked LCD (and x,y = and y,x)
	or.b		d2,d4		;LCD + new gfx
	move.b		d4,(a2)+	;Plot
	dbra.s		d7,\col

\nxt	lsl.w		#8,d1		;Prepare for last gfx
	lsl.w		#8,d5		;Prepare for last mask
\sm2	move.b		(a0),d1		;Self-modifying code: last byte to plot
\sm21	move.b		(a6),d5		;Self-modifying code: last byte to plot
	lsr.w		d0,d1		;Shift gfx
	lsr.w		d0,d5		;Shift mask

	swap		d6			;Swap in right mask into d6.lw
	and.b		d6,d5		;Mask + right mask

	and.b		d5,d1		;Mask last gfx

	move.b		(a2),d7		;LCD
	not.b		d5		;Invert gfx
	and.b		d5,d7		;Mask LCD
	or.b		d7,d1		;LCD + new gfx
	move.b		d1,(a2)		;Plot

	add.w		a4,a0		;Next row on source
	add.w		a4,a6		;Next row on source
	add.w		a5,a2		;Next row on plot
	dbra.s		d3,\row

\exit	movem.l		(sp)+,a0-a6/d0-d7
	rts


;---------------------------------------------------------------------
; PlotText(WORD ID, WORD x, WORD y, WORD flags, WORD font,
;  BYTE *string)
;
; Function: Draw the text <string> at (x,y) in window <ID> using
;  flags <flags>, font <font> and the window's current cliparea.
;
; Flagbits:
;  0 = Dimmed text on/off
;  1 = Bold text on/off
;  2 = Underlined text on/off
;  3 = Inverted text on/off
;  4 = Italic text on/off
;
; Fonts:
;   0 = TIOS small font
;   1 = TIOS medium font
;   2 = TIOS large font
;  3+ = User-defined fonts
;---------------------------------------------------------------------
PlotText:
winlib@0010:
	movem.l		a0-a6/d0-d7,-(sp)

	move.w		4+15*4+2(sp),d6		;x
	move.w		4+15*4+4(sp),d7		;y
	move.w		4+15*4+8(sp),d5		;Font

	cmp.w		#1,d5
	beq.s		\medium
	cmp.w		#2,d5
	beq.s		\large
	cmp.w		#3,d5
	bge.s		\user

\small	lea		doorsos::font_small,a2	;*font table for small TI-OS font
	move.l		#$00040006,d5		;Font height+1
	bra.s		\varied

\medium	lea		doorsos::font_medium,a2	;*font table for medium TI-OS font
	move.l		#$00070008,d5		;Font height
	move.w		#6,d4			;Font width
	bra.s		\fixed

\large	lea		doorsos::font_large,a2	;*font table for large TI-OS font
	move.l		#$0009000A,d5		;Font height
	move.w		#8,d4			;Font width
	bra.s		\fixed

\user	nop

\varied	move.l		#$12322000,\sm1
	move.l		#$12322000,\sm2
	move.l		#$121B4E71,\sm3
	bra.s		\gotfnt

\fixed	and.l		#$0000FFFF,d4
	or.l		#$123C0000,d4
	move.l		d4,\sm1
	move.l		d4,\sm2
	move.l		d4,\sm3

\gotfnt

;Make sure that at least part of the text is within the clip area
	move.w		4+15*4+0(sp),d0
	doorsos::DEREF	d0,a3			;*window

	cmp.w		W_RIGHT_CLIP(a3),d6	;Compare right clip with x
	bgt		\done			;We're to the right of the clip area, don't plot!
	cmp.w		W_BOTTOM_CLIP(a3),d7	;Compare lower clip with y
	bgt		\done			;We're below the clip area, don't plot!
	move.w		d7,d2
	add.w		d5,d2			;Below lower edge of text
	cmp.w		W_TOP_CLIP(a3),d2	;Compare upper clip with lower edge of text
	blt		\done			;We're above the clip area, don't plot!

	move.l		4+15*4+10(sp),a0		;*string

;Find first visible letter
	sub.w		W_LEFT_CLIP(a3),d6	;To make searching easier
	add.w		d5,d6
	bge.s		\xpos		;x starts to the right of x1_clip, no need to alter
	moveq		#0,d1
\findl	move.b		(a0)+,d2
	beq		\done			;Entire string is to the left of clip, don't plot!
	and.w		#$00FF,d2
	mulu.w		d5,d2
\sm1	move.b		0(a2,d2.w),d1
	add.w		d1,d6
	blt.s		\findl		;Still not visible, next letter

	sub.w		#1,a0		;Found first visible letter, go back to it
	sub.w		d1,d6		;In x too

\xpos	sub.w		d5,d6
	add.w		W_LEFT_CLIP(a3),d6	;Add back x1_clip

;Now we have enough info to set the parameters to pass to the gfx plotter
	move.l		#buffer_m,-(sp)		;*mask
	move.l		#buffer,-(sp)		;*gfx
	move.w		d7,-(sp)		;y
	move.w		d6,-(sp)		;x
	move.w		4+12+15*4+0(sp),-(sp)	;WindowID


;Find last visible letter
	move.l		a0,a1
	move.w		d6,d3
	moveq		#0,d4		;No. of visible letters

	moveq		#0,d3
	sub.w		W_RIGHT_CLIP(a3),d3	;To make searching easier
	sub.w		#12,d3

	moveq		#0,d1
\findr	move.b		(a1)+,d2
	beq		\fits		;Entire string fits within clipping area.
	addq.w		#1,d4
	and.w		#$00FF,d2
	mulu.w		d5,d2
\sm2	move.b		0(a2,d2.w),d1
	add.w		d1,d3
	blt.s		\findr		;Still visible, next letter

\fits	add.w		W_RIGHT_CLIP(a3),d3	;Add back x2_clip
	add.w		#12,d3

	move.w		d3,a6		;Width (in pixels) of visible letters
	lsr.w		#3,d3
	move.w		d3,a5
	addq.w		#1+1,a5		;Buffer width (extra space needed in case of bold/italics)


	move.w		d5,d7
	swap		d5

;Find and clear buffer
	move.l		#buffer,a1
	move.w		a5,(a1)+		;Set buffer width
	move.w		d5,(a1)			;Set buffer height
	add.w		#1,(a1)+
	move.l		a1,a4
	move.w		#BUFFER_WIDTH*BUFFER_HEIGHT/2-1,d0
\clr	clr.w		buffer_m-buffer-4(a4)	;Clear mask buffer
	clr.w		(a4)+			;Clear gfx buffer
	dbra.s		d0,\clr

;Plot the characters into the buffer
	moveq		#8,d0		;Initial shift amount
	subq.w		#1,d4
\char	move.b		(a0)+,d1
	and.w		#$00FF,d1
	mulu.w		d7,d1
	lea		0(a2,d1.w),a3
\sm3	move.b		(a3)+,d1	;Letter width in pixels
	nop

	move.l		a1,a4
	move.w		d5,d2
\plot	move.b		(a3)+,d3
	and.w		#$00FF,d3
	lsl.w		d0,d3
	or.b		d3,1(a4)
	lsr.w		#8,d3
	or.b		d3,(a4)
	add.w		a5,a4		;Next row in buffer
	dbra.s		d2,\plot	;Next letter gfx byte

	sub.b		d1,d0
	bgt.s		\next		;Still within the byte, next letter

;Next byte, create mask for this one
	move.w		d5,d2
	lea		buffer_m-buffer-4(a1),a4	;*mask buffer
\plotm	move.b		#$FF,(a4)
	add.w		a5,a4
	dbra.s		d2,\plotm

	addq.b		#8,d0
	addq.l		#1,a1

\next	dbra.s		d4,\char	;Next letter

;Create mask for the last byte
	lea		masks(pc),a3
	neg.w		d0
	move.b		8+8(a3,d0.w),d0
	move.w		d5,d2
	lea		buffer_m-buffer-4(a1),a4	;*mask buffer
\plotlm	move.b		d0,(a4)
	add.w		a5,a4
	dbra.s		d2,\plotlm



	subq.w		#2,a5

;Post-processing
	move.w		4+14+15*4+6(sp),d7	;Flags

	move.w		a5,d4
;	lsr.w		#3,d4
	subq.w		#1,d4		;Whole visible bytes
	lea		masks(pc),a0
	move.w		a6,d0
	and.w		#%0000000000000111,d0
	move.b		8(a0,d0.w),d6	;Mask for last visible byte

;Flagbits:
; 0 = Dimmed text on/off
; 1 = Bold text on/off
; 2 = Underlined text on/off
; 3 = Inverted text on/off
; 4 = Italic text on/off

;Dimmed?
	btst.b		#0,d7
	beq.s		\bold

	move.l		#buffer+4,a0
	move.b		#%01010101,d2
	move.w		d5,d0
\rowd	move.w		a5,d1
\cold	and.b		d2,(a0)+
	dbra.s		d1,\cold
	addq.l		#1,a0
	ror.b		#1,d2
	dbra.s		d0,\rowd

;Bold?
\bold	btst.b		#1,d7
	beq.s		\uline

	move.l		#buffer+4,a0
	move.w		d5,d0
\rowb	move.w		a5,d1
\colb	move.b		(a0),d2
	roxr.b		#1,d2
	or.b		d2,(a0)+
	dbra.s		d1,\colb
	addq.l		#1,a0
	dbra.s		d0,\rowb

;Underlined?
\uline	btst.b		#2,d7
	beq.s		\invert

	add.w		#1,buffer+2

	move.l		#buffer+4,a0
	move.w		a5,d0
	addq.w		#2,d0
	move.w		d5,d1
	addq.w		#1,d1
	mulu.w		d1,d0
	add.w		d0,a0

	move.w		d4,d0
	blt.s		\lastu
\loopu	move.b		#%11111111,(a0)+
	dbra.s		d0,\loopu
\lastu	move.b		d6,(a0)

;Inverted?
\invert	btst.b		#3,d7
	beq.s		\italic

	move.l		#buffer+4,a0
	move.w		d5,d0
\rowin	move.w		d4,d1
	blt.s		\lastin
\colin	not.b		(a0)+
	dbra.s		d1,\colin
\lastin	eor.b		d6,(a0)+
	addq.l		#1,a0
	dbra.s		d0,\rowin

;Italics?
\italic	btst.b		#4,d7
	beq.s		\noit

	move.l		#buffer+4,a0
	move.l		#buffer_m,a1
	move.w		d5,d0
\rowi	move.w		a5,d1
	moveq		#0,d3
	moveq		#0,d7
	move.w		d0,d4
	lsr.w		#1,d4

\coli	move.b		(a0),d2		;Read gfx
	and.w		#$00FF,d2
	ror.w		d4,d2		;Rotate gfx
	or.b		d3,d2		;Add gfx from previous byte
	move.b		d2,(a0)+	;Store new gfx
	lsr.w		#8,d2		;Get new old gfx
	move.b		d2,d3		;Store it

	move.b		(a1),d2		;Read mask
	and.w		#$00FF,d2
	ror.w		d4,d2		;Rotate mask
	or.b		d7,d2		;Add mask from previous byte
	move.b		d2,(a1)+	;Store new mask
	lsr.w		#8,d2		;Get new old mask
	move.b		d2,d7		;Store it

	dbra.s		d1,\coli
	move.b		d3,(a0)+	;Store last gfx
	move.b		d7,(a1)+	;Store last mask
	dbra.s		d0,\rowi

\noit	bsr		PlotGFX
	add.w		#14,sp

\done	movem.l		(sp)+,a0-a6/d0-d7
	rts









	include		"winlib_s.asm"


;***************************************************************************

clopen	dc.w	1,5
	dc.b	%11111110
	dc.b	%11010110
	dc.b	%11101110
	dc.b	%11010110
	dc.b	%11111110
clclose	dc.w	1,5
	dc.b	%10000010
	dc.b	%10101010
	dc.b	%10010010
	dc.b	%10101010
	dc.b	%10000010
clmask	dc.b	%11111110
	dc.b	%11111110
	dc.b	%11111110
	dc.b	%11111110
	dc.b	%11111110

words	dc.w	%1000000000000000
	dc.w	%0100000000000000
	dc.w	%0010000000000000
	dc.w	%0001000000000000
	dc.w	%0000100000000000
	dc.w	%0000010000000000
	dc.w	%0000001000000000
	dc.w	%0000000100000000
	dc.w	%0000000010000000
	dc.w	%0000000001000000
	dc.w	%0000000000100000
	dc.w	%0000000000010000
	dc.w	%0000000000001000
	dc.w	%0000000000000100
	dc.w	%0000000000000010
	dc.w	%0000000000000001

masks	dc.b	%11111111	;Lefthand masks
	dc.b	%01111111
	dc.b	%00111111
	dc.b	%00011111
	dc.b	%00001111
	dc.b	%00000111
	dc.b	%00000011
	dc.b	%00000001

	dc.b	%10000000	;Righthand masks
	dc.b	%11000000
	dc.b	%11100000
	dc.b	%11110000
	dc.b	%11111000
	dc.b	%11111100
	dc.b	%11111110
	dc.b	%11111111

	dc.b	%10000000	;Righthand masks
	dc.b	%11000000
	dc.b	%11100000
	dc.b	%11110000
	dc.b	%11111000
	dc.b	%11111100
	dc.b	%11111110
	dc.b	%00000000

wmasks	dc.w	%1111111111111111	;Lefthand masks
	dc.w	%0111111111111111
	dc.w	%0011111111111111
	dc.w	%0001111111111111
	dc.w	%0000111111111111
	dc.w	%0000011111111111
	dc.w	%0000001111111111
	dc.w	%0000000111111111
	dc.w	%0000000011111111
	dc.w	%0000000001111111
	dc.w	%0000000000111111
	dc.w	%0000000000011111
	dc.w	%0000000000001111
	dc.w	%0000000000000111
	dc.w	%0000000000000011
	dc.w	%0000000000000001

	dc.w	%1000000000000000	;Righthand masks
	dc.w	%1100000000000000
	dc.w	%1110000000000000
	dc.w	%1111000000000000
	dc.w	%1111100000000000
	dc.w	%1111110000000000
	dc.w	%1111111000000000
	dc.w	%1111111100000000
	dc.w	%1111111110000000
	dc.w	%1111111111000000
	dc.w	%1111111111100000
	dc.w	%1111111111110000
	dc.w	%1111111111111000
	dc.w	%1111111111111100
	dc.w	%1111111111111110
	dc.w	%1111111111111111
